import copy
import gc
import logging
import time
import traceback
from datetime import datetime

import numpy as np
import pandas as pd
import plotly.express as px
from sklearn.metrics import (
    accuracy_score,
    classification_report,
    matthews_corrcoef,
    mean_absolute_error,
    mean_squared_error,
    r2_score,
    recall_score,
)

from e2eml.full_processing import postprocessing

pd.set_option("display.max_colwidth", None)


class TimeTravel:
    def call_preprocessing_functions_mapping(self, class_instance):
        """
        A static function, that stores preprocessing step names, function objects and default arguments in a dictionbary.
        The order of items determines the order of preprocessing steps. We do not guarantee 100% success rate, if
        the order gets exchanged.

        :param class_instance:
        :return: Adds a class attribute.
        """
        class_instance.preprocessing_funcs = {
            "automatic_type_detection_casting": {
                "func": class_instance.automatic_type_detection_casting,
                "args": None,
            },
            "remove_duplicate_column_names": {
                "func": class_instance.remove_duplicate_column_names,
                "args": None,
            },
            "reset_dataframe_index": {
                "func": class_instance.reset_dataframe_index,
                "args": None,
            },
            "fill_infinite_values": {
                "func": class_instance.fill_infinite_values,
                "args": None,
            },
            "early_numeric_only_feature_selection": {
                "func": class_instance.automated_feature_selection,
                "args": (None, None, True),
            },
            "delete_high_null_cols": {
                "func": class_instance.delete_high_null_cols,
                "args": (0.05),
            },
            "data_binning": {"func": class_instance.data_binning, "args": None},
            "regex_clean_text_data": {
                "func": class_instance.regex_clean_text_data,
                "args": None,
            },
            "handle_target_skewness": {
                "func": class_instance.target_skewness_handling,
                "args": ("fit"),
            },
            "datetime_converter": {
                "func": class_instance.datetime_converter,
                "args": ("all"),
            },
            "pos_tagging_pca": {
                "func": class_instance.pos_tagging_pca,
                "args": (True),
            },  # slow with many categories
            "append_text_sentiment_score": {
                "func": class_instance.append_text_sentiment_score,
                "args": None,
            },
            "tfidf_vectorizer_to_pca": {
                "func": class_instance.tfidf_vectorizer_to_pca,
                "args": (True),
            },  # slow with many categories
            "tfidf_vectorizer": {
                "func": class_instance.tfidf_vectorizer_to_pca,
                "args": (False),
            },
            "rare_feature_processing": {
                "func": class_instance.rare_feature_processor,
                "args": (0.005, "miscellaneous", class_instance.rarity_cols),
            },
            "cardinality_remover": {
                "func": class_instance.cardinality_remover,
                "args": (100),
            },
            "holistic_null_filling": {
                "func": class_instance.holistic_null_filling,
                "args": (False),
            },  # slow
            "numeric_binarizer_pca": {
                "func": class_instance.numeric_binarizer_pca,
                "args": None,
            },
            "onehot_pca": {"func": class_instance.onehot_pca, "args": None},
            "category_encoding": {
                "func": class_instance.category_encoding,
                "args": ("target"),
            },
            "fill_nulls_static": {
                "func": class_instance.fill_nulls,
                "args": ("static"),
            },
            "outlier_care": {
                "func": class_instance.outlier_care,
                "args": ("isolation", "append"),
            },
            "delete_outliers": {
                "func": class_instance.outlier_care,
                "args": ("isolation", "delete", -0.5),
            },
            "remove_collinearity": {
                "func": class_instance.remove_collinearity,
                "args": (0.8),
            },
            "skewness_removal": {
                "func": class_instance.skewness_removal,
                "args": (False),
            },
            "automated_feature_transformation": {
                "func": class_instance.automated_feature_transformation,
                "args": None,
            },
            "clustering_as_a_feature_dbscan": {
                "func": class_instance.dbscan_clustering,
                "args": None,
            },
            "clustering_as_a_feature_kmeans_loop": {
                "func": class_instance.kmeans_clustering_loop,
                "args": None,
            },
            "clustering_as_a_feature_gaussian_mixture_loop": {
                "func": class_instance.gaussian_mixture_clustering_loop,
                "args": None,
            },
            "pca_clustering_results": {
                "func": class_instance.pca_clustering_results,
                "args": None,
            },
            "autotuned_clustering": {
                "func": class_instance.auto_tuned_clustering,
                "args": None,
            },
            "svm_outlier_detection_loop": {
                "func": class_instance.svm_outlier_detection_loop,
                "args": None,
            },
            "reduce_memory_footprint": {
                "func": class_instance.reduce_memory_footprint,
                "args": None,
            },
            "scale_data": {"func": class_instance.data_scaling, "args": None},
            "smote": {"func": class_instance.smote_binary_multiclass, "args": None},
            "automated_feature_selection": {
                "func": class_instance.automated_feature_selection,
                "args": (None, None, False),
            },
            "bruteforce_random_feature_selection": {
                "func": class_instance.bruteforce_random_feature_selection,
                "args": None,
            },
            "synthetic_data_augmentation": {
                "func": class_instance.synthetic_data_augmentation,
                "args": None,
            },
            "final_pca_dimensionality_reduction": {
                "func": class_instance.final_pca_dimensionality_reduction,
                "args": None,
            },
            "final_kernel_pca_dimensionality_reduction": {
                "func": class_instance.final_kernel_pca_dimensionality_reduction,
                "args": None,
            },
            "delete_low_variance_features": {
                "func": class_instance.delete_low_variance_features,
                "args": None,
            },
            "shap_based_feature_selection": {
                "func": class_instance.shap_based_feature_selection,
                "args": None,
            },
            "autoencoder_based_oversampling": {
                "func": class_instance.autoencoder_based_oversampling,
                "args": None,
            },
            "random_trees_embedding": {
                "func": class_instance.random_trees_embedding,
                "args": None,
            },  # slow
            "delete_unpredictable_training_rows": {
                "func": class_instance.delete_unpredictable_training_rows,
                "args": None,
            },
            "sort_columns_alphabetically": {
                "func": class_instance.sort_columns_alphabetically,
                "args": None,
            },
        }

    def call_classification_algorithm_mapping(self, class_instance):
        """
        A static function, that stores ml algorithms and their function objects..
        The order of items does not have any impact.

        :param class_instance: e2eml ClassificationBlueprint or RegressionBlueprint class instance.
        :return: Adds a class attribute.
        """
        class_instance.classification_algorithms_functions = {
            "logistic_regression": class_instance.ml_bp00_train_test_binary_full_processing_log_reg_prob,
            "ridge": class_instance.ml_bp08_multiclass_full_processing_ridge,
            "catboost": class_instance.ml_bp09_multiclass_full_processing_catboost,
            "xgboost": class_instance.ml_bp01_multiclass_full_processing_xgb_prob,
            "ngboost": class_instance.ml_bp04_multiclass_full_processing_ngboost,
            "lgbm": class_instance.ml_bp02_multiclass_full_processing_lgbm_prob,
            "lgbm_focal": class_instance.ml_bp14_multiclass_full_processing_lgbm_focal,
            "tabnet": class_instance.ml_bp07_multiclass_full_processing_tabnet,
            "vowpal_wabbit": class_instance.ml_bp05_multiclass_full_processing_vowpal_wabbit,
            "sklearn_ensemble": class_instance.ml_bp03_multiclass_full_processing_sklearn_stacking_ensemble,
            "sgd": class_instance.ml_bp10_multiclass_full_processing_sgd,
            "quadratic_discriminant_analysis": class_instance.ml_bp11_multiclass_full_processing_quadratic_discriminant_analysis,
            "svm": class_instance.ml_bp12_multiclass_full_processing_svm,
            "multinomial_nb": class_instance.ml_bp13_multiclass_full_processing_multinomial_nb,
            "deesc": class_instance.ml_bp15_multiclass_full_processing_deesc,
        }

    def call_regression_algorithm_mapping(self, class_instance):
        class_instance.regression_algorithms_functions = {
            "linear_regression": class_instance.ml_bp10_train_test_regression_full_processing_linear_reg,
            "elasticnet": class_instance.ml_bp19_regression_full_processing_elasticnet_reg,
            "ridge": class_instance.ml_bp18_regression_full_processing_ridge_reg,
            "catboost": class_instance.ml_bp20_regression_full_processing_catboost,
            "xgboost": class_instance.ml_bp11_regression_full_processing_xgboost,
            "ngboost": class_instance.ml_bp14_regressions_full_processing_ngboost,
            "lgbm": class_instance.ml_bp12_regressions_full_processing_lgbm,
            "tabnet": class_instance.ml_bp17_regression_full_processing_tabnet_reg,
            "vowpal_wabbit": class_instance.ml_bp15_regression_full_processing_vowpal_wabbit_reg,
            "sklearn_ensemble": class_instance.ml_bp13_regression_full_processing_sklearn_stacking_ensemble,
            "sgd": class_instance.ml_bp20_regression_full_processing_sgd,
            "svm_regression": class_instance.ml_bp22_regression_full_processing_svm
            # "ransac": class_instance.ml_bp21_regression_full_processing_ransac
        }

    def create_time_travel_checkpoints(
        self, class_instance, checkpoint_file_path=None, df=None, reload_instance=False
    ):
        """
        Runs a preprocessing blueprint only. Saves blueprints after certain checkpoints, which can be defined in

        :param class_instance: Accepts a an e2eml Classification or Regression class instance. This does not support
            NLP transformers.
        :param checkpoint_file_path: (Optional). Takes a file path to store the saved class instance checkpoints.
            On default will save in current location.
        :param df: Accepts a dataframe to make predictions on new data.
        :return: Saves the checkpoints locally.
        """
        if not reload_instance:
            logging.info("Start blueprint.")
            class_instance.runtime_warnings(warn_about="future_architecture_change")
            class_instance.check_prediction_mode(df)
            class_instance.train_test_split(how=class_instance.train_split_type)
            self.last_checkpoint_reached = "train_test_split"
        else:
            pass

        self.call_preprocessing_functions_mapping(class_instance=class_instance)
        if class_instance.class_problem in ["binary", "multiclass"]:
            self.call_classification_algorithm_mapping(class_instance=class_instance)
        else:
            self.call_regression_algorithm_mapping(class_instance=class_instance)

        for key in class_instance.blueprint_step_selection_non_nlp.keys():
            if class_instance.blueprint_step_selection_non_nlp[key] and (
                not class_instance.checkpoint_reached[key]
                or class_instance.prediction_mode
            ):
                if (
                    (
                        key == "regex_clean_text_data"
                        and len(class_instance.nlp_transformer_columns) > 0
                    )
                    or (
                        key == "tfidf_vectorizer"
                        and len(class_instance.nlp_transformer_columns) > 0
                    )
                    or (
                        key == "append_text_sentiment_score"
                        and len(class_instance.nlp_transformer_columns) > 0
                    )
                    or (
                        key
                        not in [
                            "regex_clean_text_data",
                            "tfidf_vectorizer",
                            "append_text_sentiment_score",
                            "train_test_split",
                        ]
                    )
                ):
                    if class_instance.preprocessing_funcs[key]["args"]:
                        try:
                            if (
                                len(
                                    np.array(
                                        class_instance.preprocessing_funcs[key]["args"]
                                    )
                                )
                                == 1
                            ):
                                class_instance.preprocessing_funcs[key]["func"](
                                    class_instance.preprocessing_funcs[key]["args"]
                                )
                            else:
                                class_instance.preprocessing_funcs[key]["func"](
                                    *class_instance.preprocessing_funcs[key]["args"]
                                )
                        except TypeError:
                            class_instance.preprocessing_funcs[key]["func"](
                                class_instance.preprocessing_funcs[key]["args"]
                            )
                    else:
                        class_instance.preprocessing_funcs[key]["func"]()
                else:
                    print(
                        f"Skipped preprocessing step {key} as it has not been selected by user."
                    )

                # save checkpoints, if not in predict
                if class_instance.prediction_mode:
                    pass
                else:
                    class_instance.checkpoint_reached[key] = True
                    self.last_checkpoint_reached = key
                    postprocessing.save_to_production(
                        class_instance,
                        file_name=f"blueprint_checkpoint_{key}",
                        clean=False,
                        file_path=checkpoint_file_path,
                    )
            else:
                pass

    def load_checkpoint(self, checkpoint_to_load=None, checkpoint_file_path=None):
        """
        This function loads saved checkpoints on demand. If no checkpoint is specified, it loads the most recent
        checkpoint executed in the pipeline.

        :param checkpoint_to_load: Takes a string, specifying the checkpoint to load. All strings can be looked up
            in blueprint_step_selection_non_nlp class attribute. Only checkpoints, which have been saved explicitely can be
            loaded. If no checkpoint is specified, it loads the most recent
            checkpoint executed in the pipeline.
        :param checkpoint_file_path: (Optional) Takes a string. On default loads checkpoint from current path. If specified,
            loads the checkpoint from this path.
        :return: Returns loaded checkpoint.
        """
        if not checkpoint_to_load:
            class_instance = postprocessing.load_for_production(
                file_name=f"blueprint_checkpoint_{self.last_checkpoint_reached}",
                file_path=checkpoint_file_path,
            )
        else:
            class_instance = postprocessing.load_for_production(
                file_name=f"blueprint_checkpoint_{checkpoint_to_load}",
                file_path=checkpoint_file_path,
            )
        return class_instance

    def timetravel_model_training(self, class_instance, algorithm="lgbm"):
        def overwrite_normal_preprocess_func(df=None):
            return "overwritten"

        # overwriting the preprocessing function as we have done this already
        class_instance.std_preprocessing_pipeline = overwrite_normal_preprocess_func

        if class_instance.class_problem in ["binary", "multiclass"]:
            class_instance.classification_algorithms_functions[algorithm]()
        else:
            class_instance.regression_algorithms_functions[algorithm]()

        logging.info("Finished blueprint.")


def timewalk_auto_exploration(  # noqa: C901
    class_instance,
    holdout_df,
    holdout_target,
    is_imbalanced=False,
    algs_to_test=None,
    preprocess_checkpoints=None,
    speed_up_model_tuning=True,
    name_of_exist_experiment=None,
    experiment_name="timewalk.pkl",
    experiment_comment=f'Experiment run at {datetime.now().strftime("%Y-%m-%d %H:%M:%S")}',
):
    """
    Timewalk is an extension to TimeTravel. It executes different preprocessing steps an short model training to explore
    the best combination. It returns a Pandas DataFrame with all results. Timewalk is meant to explore and is not suitable
    for final training for various reasons. Timewalk will result in long runtimes.

    :param class_instance: Expects a freshly instantiated e2eml ClassificationBlueprint or RegressionBlueprint class instance.
    :param holdout_df: Expects a Pandas Dataframe, which will be only used for final evaluation. Is not alolowed to
        include the target as it simulates prediction on new data. It is recommended to use the most recent datapoints as holdout.
    :param holdout_target: Expects a Pandas Series with holdout targets.
    :param speed_up_model_tuning: If True, timewalk will run with reduced rounds of hyprparameter tuning. If False,
        timewalk will use hyperparameter tuning rounds and maximum runtime from the imported e2eml class. If False, users can
        control these parameter by adjusting them after e2eml class instantiation and before importing the class to timewalk.
    :param name_of_exist_experiment: Expects a string. Name of a locally saved file with results from a past experiment
        can be provided. In this case timewalk will load the file as a dataframe and concatenate old and new results into
        one dataframe.
    :param algs_to_test: (Optional). Expects a list object with algorithms to test. Will test on default:
        ["ridge", "xgboost", "lgbm", "tabnet", "ngboost", "vowpal_wabbit", "logistic_regression",
        "linear_regression", "elasticnet", "sgd", "quadratic_discriminant_analysis", "svm"]
    :param experiment_name: Expects string. Will determine the name of the exported results dataframe (as pickle file).
    :param experiment_comment: Expects a string. This will add a comment of choice to the results dataframe. On default
        it will add a string stating when the experiment has been started.
    :return: Pandas DataFrame with results.
    """
    class_instance = copy.copy(class_instance)

    # define algorithms to consider
    if isinstance(algs_to_test, list):
        algorithms = algs_to_test
    else:

        algorithms = [
            "ridge",
            "xgboost",
            "lgbm",
            "lgbm_focal",
            "tabnet",
            "ngboost",
            "vowpal_wabbit",
            "logistic_regression",
            "linear_regression",
            "elasticnet",
            "sgd",
            "quadratic_discriminant_analysis",
            "svm",
            "svm_regression",
        ]

        if len(class_instance.dataframe.index) > 10000:
            algorithms.remove("ngboost")
            algorithms.remove("xgboost")
        else:
            pass

    # removing algorithms, that are not suitable for classification or regression tasks
    if class_instance.class_problem in ["binary", "multiclass"]:
        try:
            algorithms.remove("linear_regression")
        except Exception:
            pass

        try:
            algorithms.remove("elasticnet")
        except Exception:
            pass

        try:
            algorithms.remove("ransac")
        except Exception:
            pass

        try:
            algorithms.remove("svm_regression")
        except Exception:
            pass

    else:
        try:
            algorithms.remove("logistic_regression")
        except Exception:
            pass

        try:
            algorithms.remove("quadratic_discriminant_analysis")
        except Exception:
            pass

        try:
            algorithms.remove("svm")
        except Exception:
            pass

        try:
            algorithms.remove("multinomial_nb")
        except Exception:
            pass

        try:
            algorithms.remove("lgbm_focal")
        except Exception:
            pass

    if speed_up_model_tuning:
        # we reduce the tuning rounds for all algorithms
        class_instance.hyperparameter_tuning_rounds["xgboost"] = 10
        class_instance.hyperparameter_tuning_rounds["lgbm"] = 10
        class_instance.hyperparameter_tuning_rounds["sgd"] = 100
        class_instance.hyperparameter_tuning_rounds["svm"] = 25
        class_instance.hyperparameter_tuning_rounds["svm_regression"] = 10
        class_instance.hyperparameter_tuning_rounds["multinomial_nb"] = 10
        class_instance.hyperparameter_tuning_rounds["tabnet"] = 10
        class_instance.hyperparameter_tuning_rounds["ngboost"] = 3
        class_instance.hyperparameter_tuning_rounds["sklearn_ensemble"] = 10
        class_instance.hyperparameter_tuning_rounds["ridge"] = 10
        class_instance.hyperparameter_tuning_rounds["elasticnet"] = 10
        class_instance.hyperparameter_tuning_rounds["catboost"] = 3
        class_instance.hyperparameter_tuning_rounds["bruteforce_random"] = 50
        class_instance.hyperparameter_tuning_rounds[
            "autoencoder_based_oversampling"
        ] = 10
        class_instance.hyperparameter_tuning_rounds[
            "final_pca_dimensionality_reduction"
        ] = 50

        # we also limit the time for hyperparameter tuning
        class_instance.hyperparameter_tuning_max_runtime_secs["xgboost"] = 4 * 60 * 60
        class_instance.hyperparameter_tuning_max_runtime_secs["sgd"] = 3 * 60 * 60
        class_instance.hyperparameter_tuning_max_runtime_secs["lgbm"] = 3 * 60 * 60
        class_instance.hyperparameter_tuning_max_runtime_secs["tabnet"] = 2 * 60 * 60
        class_instance.hyperparameter_tuning_max_runtime_secs["ngboost"] = 3 * 60 * 60
        class_instance.hyperparameter_tuning_max_runtime_secs["sklearn_ensemble"] = (
            3 * 60 * 60
        )
        class_instance.hyperparameter_tuning_max_runtime_secs["ridge"] = 3 * 60 * 60
        class_instance.hyperparameter_tuning_max_runtime_secs["elasticnet"] = (
            3 * 60 * 60
        )
        class_instance.hyperparameter_tuning_max_runtime_secs["catboost"] = 4 * 60 * 60
        class_instance.hyperparameter_tuning_max_runtime_secs["bruteforce_random"] = (
            3 * 60 * 60
        )
        class_instance.hyperparameter_tuning_max_runtime_secs[
            "autoencoder_based_oversampling"
        ] = (1 * 60 * 60)
        class_instance.hyperparameter_tuning_max_runtime_secs[
            "final_pca_dimensionality_reduction"
        ] = (1 * 60 * 60)
    else:
        pass

    # we adjust default preprocessing
    class_instance.blueprint_step_selection_non_nlp["autotuned_clustering"] = True
    class_instance.blueprint_step_selection_non_nlp["scale_data"] = True

    # we want to store our results
    scoring_results = []
    scoring_2_results = []
    scoring_3_results = []
    # scoring_4_results = []
    algorithms_used = []
    preprocessing_steps_used = []
    elapsed_times = []
    unique_indices = []

    # define checkpoints to load
    if isinstance(preprocess_checkpoints, list):
        checkpoints = preprocess_checkpoints
    else:
        checkpoints = [
            "default",
            "delete_low_variance_features",
            "automated_feature_selection",
            "autotuned_clustering",
            "cardinality_remover",
            "delete_high_null_cols",
            "early_numeric_only_feature_selection",
            "fill_infinite_values",
        ]

    if not is_imbalanced:
        if "shap_based_feature_selection" in checkpoints:
            checkpoints.remove("shap_based_feature_selection")

    # define the type of scoring
    if class_instance.class_problem in ["binary", "multiclass"]:
        metric = "Matthews"
        metric_2 = "Accuracy"
        metric_3 = "Recall"
        # ascending = False
    else:
        metric = "Mean absolute error"
        metric_2 = "R2 score"
        metric_3 = "RMSE"
        # ascending = True

    # creating checkpoints and training the model
    automl_travel = TimeTravel()

    unique_indices_counter = 0
    for checkpoint in checkpoints:
        preprocess_runtime = 0
        preprocess_start = time.time()
        print("+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
        print(f"Start iteration for checkpoint {checkpoint} at {preprocess_start}.")
        try:
            if len(scoring_results) == 0:
                automl_travel.create_time_travel_checkpoints(
                    class_instance, reload_instance=False
                )
            else:
                class_instance = automl_travel.load_checkpoint(
                    checkpoint_to_load=checkpoint
                )
                if checkpoint == "delete_low_variance_features":
                    class_instance.blueprint_step_selection_non_nlp[
                        "autoencoder_based_oversampling"
                    ] = True
                    class_instance.blueprint_step_selection_non_nlp[
                        "shap_based_feature_selection"
                    ] = False
                    class_instance.blueprint_step_selection_non_nlp[
                        "autotuned_clustering"
                    ] = False
                elif checkpoint == "automated_feature_selection":
                    class_instance.blueprint_step_selection_non_nlp[
                        "autoencoder_based_oversampling"
                    ] = False
                    class_instance.blueprint_step_selection_non_nlp[
                        "shap_based_feature_selection"
                    ] = False
                    class_instance.blueprint_step_selection_non_nlp[
                        "autotuned_clustering"
                    ] = False
                elif checkpoint == "autotuned_clustering":
                    class_instance.blueprint_step_selection_non_nlp[
                        "shap_based_feature_selection"
                    ] = True
                    class_instance.blueprint_step_selection_non_nlp[
                        "autotuned_clustering"
                    ] = False
                    class_instance.blueprint_step_selection_non_nlp[
                        "delete_unpredictable_training_rows"
                    ] = True
                    class_instance.blueprint_step_selection_non_nlp[
                        "autotuned_clustering"
                    ] = True
                elif checkpoint == "cardinality_remover":
                    class_instance.blueprint_step_selection_non_nlp[
                        "autoencoder_based_oversampling"
                    ] = False
                    class_instance.blueprint_step_selection_non_nlp[
                        "svm_outlier_detection_loop"
                    ] = False
                    class_instance.blueprint_step_selection_non_nlp[
                        "autotuned_clustering"
                    ] = False
                    class_instance.blueprint_step_selection_non_nlp[
                        "shap_based_feature_selection"
                    ] = True
                    class_instance.blueprint_step_selection_non_nlp[
                        "delete_outliers"
                    ] = True
                    class_instance.blueprint_step_selection_non_nlp[
                        "automated_feature_transformation"
                    ] = True
                elif checkpoint == "delete_high_null_cols":
                    class_instance.blueprint_step_selection_non_nlp[
                        "autotuned_clustering"
                    ] = False
                    class_instance.blueprint_step_selection_non_nlp[
                        "tfidf_vectorizer_to_pca"
                    ] = True
                    class_instance.blueprint_step_selection_non_nlp[
                        "data_binning"
                    ] = True
                    class_instance.blueprint_step_selection_non_nlp[
                        "svm_outlier_detection_loop"
                    ] = False
                    class_instance.blueprint_step_selection_non_nlp[
                        "autoencoder_based_oversampling"
                    ] = False
                    class_instance.blueprint_step_selection_non_nlp[
                        "delete_outliers"
                    ] = False
                    class_instance.blueprint_step_selection_non_nlp[
                        "automated_feature_transformation"
                    ] = False
                    class_instance.blueprint_step_selection_non_nlp[
                        "shap_based_feature_selection"
                    ] = False
                elif checkpoint == "early_numeric_only_feature_selection":
                    class_instance.blueprint_step_selection_non_nlp[
                        "autoencoder_based_oversampling"
                    ] = False
                    class_instance.blueprint_step_selection_non_nlp[
                        "tfidf_vectorizer_to_pca"
                    ] = False
                    class_instance.blueprint_step_selection_non_nlp[
                        "data_binning"
                    ] = False
                    class_instance.blueprint_step_selection_non_nlp[
                        "delete_outliers"
                    ] = True
                    class_instance.blueprint_step_selection_non_nlp[
                        "automated_feature_transformation"
                    ] = True
                elif checkpoint == "fill_infinite_values":
                    class_instance.blueprint_step_selection_non_nlp[
                        "tfidf_vectorizer_to_pca"
                    ] = False
                    class_instance.blueprint_step_selection_non_nlp[
                        "numeric_binarizer_pca"
                    ] = False
                    class_instance.blueprint_step_selection_non_nlp[
                        "data_binning"
                    ] = False
                    class_instance.blueprint_step_selection_non_nlp[
                        "svm_outlier_detection_loop"
                    ] = False
                    class_instance.blueprint_step_selection_non_nlp[
                        "autoencoder_based_oversampling"
                    ] = False
                    class_instance.blueprint_step_selection_non_nlp[
                        "clustering_as_a_feature_dbscan"
                    ] = False
                    class_instance.blueprint_step_selection_non_nlp[
                        "clustering_as_a_feature_kmeans_loop"
                    ] = False
                    class_instance.blueprint_step_selection_non_nlp[
                        "clustering_as_a_feature_gaussian_mixture_loop"
                    ] = False
                    class_instance.blueprint_step_selection_non_nlp[
                        "pca_clustering_results"
                    ] = False
                    class_instance.blueprint_step_selection_non_nlp[
                        "delete_outliers"
                    ] = True
                    class_instance.blueprint_step_selection_non_nlp[
                        "automated_feature_transformation"
                    ] = True
                    class_instance.blueprint_step_selection_non_nlp[
                        "shap_based_feature_selection"
                    ] = True
            automl_travel.create_time_travel_checkpoints(
                class_instance, reload_instance=True
            )

        except Exception:
            print(traceback.format_exc())
        preprocess_end = time.time()
        preprocess_runtime = preprocess_end - preprocess_start
        print(f"All algorithms to be tested are: {algorithms}.")
        try:
            for alg in algorithms:
                unique_indices_counter += 1
                start = time.time()
                print(f"Start iteration for algorithm {alg} at {start}.")
                try:
                    class_instance = automl_travel.load_checkpoint(
                        checkpoint_to_load="sort_columns_alphabetically"
                    )  # gets latest checkpoint
                    print("Successfully loaded checkpoint last checkpoint.")
                    automl_travel.create_time_travel_checkpoints(
                        class_instance, reload_instance=True
                    )
                    automl_travel.timetravel_model_training(class_instance, alg)
                    automl_travel.create_time_travel_checkpoints(
                        class_instance, df=holdout_df
                    )
                    automl_travel.timetravel_model_training(class_instance, alg)
                    if class_instance.labels_encoded:
                        hold_df_target = class_instance.label_encoder_decoder(
                            holdout_target, mode="transform"
                        )
                    else:
                        hold_df_target = holdout_target
                except Exception:
                    print(traceback.format_exc())

                try:
                    if class_instance.class_problem in ["binary", "multiclass"]:
                        scoring_2 = accuracy_score(
                            hold_df_target, class_instance.predicted_classes[alg]
                        )
                        scoring_3 = recall_score(
                            hold_df_target,
                            class_instance.predicted_classes[alg],
                            average="weighted",
                        )
                        scoring = matthews_corrcoef(
                            hold_df_target, class_instance.predicted_classes[alg]
                        )
                        full_classification_report = classification_report(
                            hold_df_target, class_instance.predicted_classes[alg]
                        )
                        print(full_classification_report)
                    else:
                        scoring = mean_absolute_error(
                            hold_df_target, class_instance.predicted_values[alg]
                        )
                        scoring_2 = r2_score(
                            hold_df_target, class_instance.predicted_values[alg]
                        )
                        scoring_3 = mean_squared_error(
                            hold_df_target,
                            class_instance.predicted_values[alg],
                            squared=True,
                        )
                except Exception:
                    print(traceback.format_exc())
                    try:
                        if class_instance.class_problem in ["binary", "multiclass"]:
                            scoring = matthews_corrcoef(
                                pd.Series(hold_df_target).astype(bool),
                                class_instance.predicted_classes[alg],
                            )
                            scoring_2 = accuracy_score(
                                pd.Series(hold_df_target).astype(bool),
                                class_instance.predicted_classes[alg].astype(bool),
                            )
                            scoring_3 = recall_score(
                                pd.Series(hold_df_target).astype(bool),
                                class_instance.predicted_classes[alg].astype(bool),
                                average="weighted",
                            )
                            full_classification_report = classification_report(
                                pd.Series(hold_df_target).astype(bool),
                                class_instance.predicted_classes[alg],
                            )
                            print(full_classification_report)
                        else:
                            scoring = mean_absolute_error(
                                pd.Series(hold_df_target).astype(float),
                                class_instance.predicted_values[alg],
                            )
                            scoring_2 = r2_score(
                                pd.Series(hold_df_target).astype(float),
                                class_instance.predicted_values[alg],
                            )
                            scoring_3 = mean_squared_error(
                                pd.Series(hold_df_target).astype(float),
                                class_instance.predicted_values[alg],
                                squared=True,
                            )
                    except Exception:
                        print(traceback.format_exc())
                        print("Evaluation on holdout failed.")
                        if class_instance.class_problem in ["binary", "multiclass"]:
                            scoring = 0
                            scoring_2 = 0
                            scoring_3 = 0
                        else:
                            scoring = 999999999
                            scoring_2 = -1
                            scoring_3 = 99999999

                print(f"Score achieved on holdout dataset is: {scoring}.")

                end = time.time()
                elapsed_time = end - start
                elapsed_times.append(elapsed_time)
                scoring_results.append(scoring)
                scoring_2_results.append(scoring_2)
                scoring_3_results.append(scoring_3)
                algorithms_used.append(alg)
                unique_indices.append(unique_indices_counter)
                preprocessing_steps_used.append(
                    class_instance.blueprint_step_selection_non_nlp
                )
                results_dict = {
                    "Trial number": unique_indices,
                    "Algorithm": algorithms_used,
                    metric: scoring_results,
                    metric_2: scoring_2_results,
                    metric_3: scoring_3_results,
                    "Preprocessing applied": preprocessing_steps_used,
                    "ML model runtime in seconds": elapsed_times,
                }

                results_df = pd.DataFrame(results_dict)
                results_df["experiment_comment"] = experiment_comment
                results_df["preprocessing_runtime_sec"] = preprocess_runtime
                results_df["Total runtime"] = (
                    results_df["ML model runtime in seconds"]
                    + results_df["preprocessing_runtime_sec"]
                )
                if isinstance(name_of_exist_experiment, str):
                    try:
                        print("Try to load former experiment data.")
                        loaded_experiment = pd.read_pickle(name_of_exist_experiment)
                        results_df = pd.concat([results_df, loaded_experiment])
                    except Exception:
                        print(traceback.format_exc())
                        print(
                            "Loading former experiment data failed. Will export running experiment only."
                        )
                else:
                    pass
                results_df.to_pickle(experiment_name)
                print(f"End iteration for algorithm {alg} at {end}.")
                print("+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
                try:
                    del class_instance.trained_models[alg]
                    del class_instance
                    _ = gc.collect()
                except KeyError:
                    pass
        except Exception:
            print(traceback.format_exc())
            if class_instance.class_problem in ["binary", "multiclass"]:  # noqa: F821
                scoring = 0
                scoring_2 = 0
                scoring_3 = 0
            else:
                scoring = 999999999
                scoring_2 = -1
                scoring_3 = 99999999
            end = time.time()
            elapsed_time = end - start
            elapsed_times.append(elapsed_time)
            scoring_results.append(scoring)
            scoring_2_results.append(scoring_2)
            scoring_3_results.append(scoring_3)
            algorithms_used.append(alg)
            unique_indices.append(unique_indices_counter)
            preprocessing_steps_used.append(  # noqa: F821
                class_instance.blueprint_step_selection_non_nlp  # noqa: F821
            )  # noqa: F821
            results_dict = {
                "Trial number": unique_indices,
                "Algorithm": algorithms_used,
                metric: scoring_results,
                metric_2: scoring_2_results,
                metric_3: scoring_3_results,
                "Preprocessing applied": preprocessing_steps_used,
                "ML model runtime in seconds": elapsed_times,
            }

            results_df = pd.DataFrame(results_dict)
            results_df["experiment_comment"] = experiment_comment
            results_df["preprocessing_runtime_sec"] = preprocess_runtime
            results_df["Total runtime"] = (
                results_df["ML model runtime in seconds"]
                + results_df["preprocessing_runtime_sec"]
            )
            if isinstance(name_of_exist_experiment, str):
                try:
                    print("Try to load former experiment data.")
                    loaded_experiment = pd.read_pickle(name_of_exist_experiment)
                    results_df = pd.concat([results_df, loaded_experiment])
                except Exception:
                    print(traceback.format_exc())
                    print(
                        "Loading former experiment data failed. Will export running experiment only."
                    )
            else:
                pass
            results_df.to_pickle(experiment_name)
            print(f"End iteration for algorithm {alg} at {end}.")
            print("+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
            try:
                del class_instance.trained_models[alg]  # noqa: F821
                del class_instance  # noqa: F821
                _ = gc.collect()
            except KeyError:
                pass

    results_dict = {
        "Trial number": unique_indices,
        "Algorithm": algorithms_used,
        metric: scoring_results,
        metric_2: scoring_2_results,
        metric_3: scoring_3_results,
        "Preprocessing applied": preprocessing_steps_used,
        "ML model runtime in seconds": elapsed_times,
    }

    results_df = pd.DataFrame(results_dict)
    results_df["experiment_comment"] = experiment_comment
    results_df["preprocessing_runtime_sec"] = preprocess_runtime
    results_df["Total runtime"] = (
        results_df["ML model runtime in seconds"]
        + results_df["preprocessing_runtime_sec"]
    )
    if isinstance(name_of_exist_experiment, str):
        try:
            print("Try to load former experiment data.")
            loaded_experiment = pd.read_pickle(name_of_exist_experiment)
            results_df = pd.concat([results_df, loaded_experiment])
        except Exception:
            print(traceback.format_exc())
            print(
                "Loading former experiment data failed. Will export running experiment only."
            )
    else:
        pass
    results_df.to_pickle(experiment_name)

    try:
        fig = px.line(
            results_df,
            x="Total runtime",
            y=metric,
            color="Algorithm",
            text="Trial number",
        )
        fig.update_traces(textposition="bottom right")
        fig.show()
    except Exception:
        print(traceback.format_exc())
    return results_df



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['e2eml',
 'e2eml.classification',
 'e2eml.full_processing',
 'e2eml.model_utils',
 'e2eml.regression',
 'e2eml.test',
 'e2eml.time_series',
 'e2eml.timetravel']

package_data = \
{'': ['*']}

install_requires = \
['boostaroota>=1.3,<2.0',
 'catboost>=0.21,<0.22',
 'category_encoders==2.3.0',
 'dill>=0.3.3,<0.4.0',
 'imbalanced-learn>=0.8,<0.9',
 'lightgbm>=3.1.0,<4.0.0',
 'matplotlib==3.1.3',
 'ngboost>=0.3.1,<0.4.0',
 'nltk>=3.2.4,<4.0.0',
 'numpy>=1.19.4,<2.0.0',
 'optuna>=2.8.0,<3.0.0',
 'pandas>=1.1.5,<2.0.0',
 'plotly>=5,<6',
 'psutil==5.8.0',
 'pytorch_tabnet>=3.1.1,<4.0.0',
 'scikit-learn>=1.0.1,<2.0.0',
 'scipy>=1.5.4,<2.0.0',
 'seaborn>=0.11.1,<0.12.0',
 'shap>=0.39.0,<0.40.0',
 'spacy>=3.1.0,<4.0.0',
 'textblob>=0,<1',
 'torch>=1.7.0,<2.0.0',
 'transformers>=4,<5',
 'vowpalwabbit>=8.11.0,<9.0.0',
 'xgboost>=1.3.3,<2.0.0']

extras_require = \
{'full': ['cupy>=8.1.0,<9.0.0',
          'cython>=0.29.21,<0.30.0',
          'ipython>=7.10.0,<8.0.0',
          'notebook>=6.1.0,<7.0.0'],
 'jupyter': ['ipython>=7.10.0,<8.0.0', 'notebook>=6.1.0,<7.0.0'],
 'rapids': ['cupy>=8.1.0,<9.0.0', 'cython>=0.29.21,<0.30.0']}

setup_kwargs = {
    'name': 'e2eml',
    'version': '2.10.3',
    'description': 'An end-to-end solution for automl',
    'long_description': '# e2e ML\n\n> An end to end solution for automl.\n\nPass in your data, add some information about it and get a full pipelines in\nreturn. Data preprocessing, feature creation, modelling and evaluation with just\na few lines of code.\n\n![Header image](header.png)\n\n## Contents\n\n<!-- toc -->\n\n* [Installation](#installation)\n* [Usage example](#usage-example)\n* [Linting and Pre-Commit](#linting-and-pre-commit)\n* [Disclaimer](#disclaimer)\n* [Development](#development)\n  * [Adding or Removing Dependencies](#adding-or-removing-dependencies)\n  * [Building and Publishing](#building-and-publishing)\n  * [Documentation](#documentation)\n  * [Pull Requests](#pull-requests)\n* [Release History](#release-history)\n* [References](#references)\n* [Meta](#meta)\n\n<!-- tocstop -->\n\n## Installation\n\nFrom PyPI:\n\n```sh\npip install e2eml\n```\n\nWe highly recommend to create a new virtual environment first. Then install\ne2e-ml into it. In the environment also download the pretrained spacy model\nwith. Otherwise e2eml will do this automatically during runtime.\n\ne2eml can also be installed into a RAPIDS environment. For this we recommend to\ncreate a fresh environment following [RAPIDS](https://rapids.ai/start.html)\ninstructions. After environment installation and activation, a special\ninstallation is needed to not run into installation issues.\n\nJust run:\n\n```sh\npip install e2eml[rapids]\n```\n\nThis will additionally install cupy and cython to prevent issues. Additionally\nit is needed to run:\n\n```sh\npip3 install torch==1.9.0+cu111 torchvision==0.10.0+cu111 torchaudio==0.9.0 -f https://download.pytorch.org/whl/torch_stable.html\n\n# also spacy supports GPU acceleration\npip install -U spacy[cuda112] #cuda112 depends on your actual cuda version, see: https://spacy.io/usage\n```\n\nOtherwise Pytorch will fail trying to run on GPU.\n\nIf e2eml shall be installed together with Jupyter core and ipython, please\ninstall with:\n\n```sh\npip install e2eml[full]\n```\n\ninstead.\n\n## Usage example\n\ne2e has been designed to create state-of-the-art machine learning pipelines with\na few lines of code. Basic example of usage:\n\n```python\nimport e2eml\nfrom e2eml.classification import classification_blueprints\nimport pandas as pd\n# import data\ndf = pd.read_csv("Your.csv")\n\n# split into a test/train & holdout set (holdout for prediction illustration here, but not required at all)\ntrain_df = df.head(1000).copy()\nholdout_df = df.tail(200).copy() # make sure\n# saving the holdout dataset\'s target for later and delete it from holdout dataset\ntarget = "target_column"\nholdout_target = holdout_df[target].copy()\ndel holdout_df[target]\n\n# instantiate the needed blueprints class\nfrom classification import classification_blueprints # regression bps are available with from regression import regression_blueprints\ntest_class = classification_blueprints.ClassificationBluePrint(datasource=train_df,\n                        target_variable=target,\n                        train_split_type=\'cross\',\n                        rapids_acceleration=True, # if installed into a conda environment with NVIDIA Rapids, this can be used to accelerate preprocessing with GPU\n                        preferred_training_mode=\'auto\', # Auto will automatically identify, if LGBM & Xgboost can use GPU acceleration*\n                        tune_mode=\'accurate\' # hyperparameter sets will be validated with 10-fold CV Set this to \'simple\' for 1-fold CV\n                        #categorical_columns=cat_columns # you can define categorical columns, otherwise e2e does this automatically\n                        #date_columns=date_columns # you can also define date columns (expected is YYYY-MM-DD format)\n                                                               )\n\n"""\n*\n\'Auto\' is recommended for preferred_training_mode parameter, but with \'CPU\' and \'GPU\' it can also be controlled manually.\nIf you install Xgboost & LGBM into the same environment as GPU accelerated versions, you can set preferred_training_mode=\'gpu\'.\nThis will massively improve training times and speed up SHAP feature importance for LGBM and Xgboost related tasks.\nFor Xgboost this should work out of the box, if installed into a RAPIDS environment.\n"""\n# run actual blueprint\ntest_class.ml_bp01_multiclass_full_processing_xgb_prob()\n\n"""\nWhen choosing blueprints several options are available:\n\nMulticlass blueprints can handle binary and multiclass tasks:\n- ml_bp00_train_test_binary_full_processing_log_reg_prob()\n- ml_bp01_multiclass_full_processing_xgb_prob()\n- ml_bp02_multiclass_full_processing_lgbm_prob()\n- ml_bp03_multiclass_full_processing_sklearn_stacking_ensemble()\n- ml_bp04_multiclass_full_processing_ngboost()\n- ml_bp05_multiclass_full_processing_vowpal_wabbit\n- ml_bp06_multiclass_full_processing_bert_transformer() # for NLP specifically\n- ml_bp07_multiclass_full_processing_tabnet()\n- ml_bp08_multiclass_full_processing_ridge()\n- ml_bp09_multiclass_full_processing_catboost()\n- ml_bp10_multiclass_full_processing_sgd()\n- ml_special_binary_full_processing_boosting_blender()\n- ml_special_multiclass_auto_model_exploration()\n- ml_special_multiclass_full_processing_multimodel_max_voting()\n\nThere are regression blueprints as well (in regression module):\n- ml_bp10_train_test_regression_full_processing_linear_reg()\n- ml_bp11_regression_full_processing_xgboost()\n- ml_bp12_regressions_full_processing_lgbm()\n- ml_bp13_regression_full_processing_sklearn_stacking_ensemble()\n- ml_bp14_regressions_full_processing_ngboost()\n- ml_bp15_regression_full_processing_vowpal_wabbit_reg()\n- ml_bp16_regressions_full_processing_bert_transformer()\n- ml_bp17_regression_full_processing_tabnet_reg()\n- ml_bp18_regression_full_processing_ridge_reg\n- ml_bp20_regression_full_processing_catboost()\n- ml_bp20_regression_full_processing_sgd()\n- ml_special_regression_full_processing_multimodel_avg_blender()\n- ml_special_regression_auto_model_exploration()\n\nIn ensembles algorithms can be chosen via the class attribute:\ntest_class.special_blueprint_algorithms = {"ridge": True,\n                                            "elasticnet": False,\n                                             "xgboost": True,\n                                             "ngboost": True,\n                                             "lgbm": True,\n                                             "tabnet": False,\n                                             "vowpal_wabbit": True,\n                                             "sklearn_ensemble": True,\n                                             "catboost": False\n                                             }\n\nAlso preprocessing steps can be selected:\ntest_class.blueprint_step_selection_non_nlp = {\n            "automatic_type_detection_casting": True,\n            "early_numeric_only_feature_selection": True,\n            "remove_duplicate_column_names": True,\n            "reset_dataframe_index": True,\n            "regex_clean_text_data": False,\n            "handle_target_skewness": False,\n            "holistic_null_filling": True, # slow\n            "iterative_null_imputation": False, # very slow\n            "fill_infinite_values": True,\n            "datetime_converter": True,\n            "pos_tagging_pca": False, # slow with many categories\n            "append_text_sentiment_score": False,\n            "tfidf_vectorizer_to_pca": True, # slow with many categories\n            "tfidf_vectorizer": False,\n            "rare_feature_processing": True,\n            "cardinality_remover": True,\n            "delete_high_null_cols": True,\n            "numeric_binarizer_pca": True,\n            "onehot_pca": True,\n            "category_encoding": True,\n            "fill_nulls_static": True,\n            "data_binning": True,\n            "outlier_care": True,\n            "remove_collinearity": True,\n            "skewness_removal": True,\n            "clustering_as_a_feature_dbscan": True,\n            "clustering_as_a_feature_kmeans_loop": True, # slow for big data, but can be heavily accelerated using rapids_acceleration=True during class instantiation\n            "clustering_as_a_feature_gaussian_mixture_loop": True, # slow for big data, but can be heavily accelerated using rapids_acceleration=True during class instantiation (will run a Kmeans on GPU)\n            "pca_clustering_results": True,\n            "reduce_memory_footprint": False,\n            "automated_feature_selection": True,\n            "bruteforce_random_feature_selection": False, # slow, this feature is experimental!\n            "sort_columns_alphabetically": True,\n            "synthetic_data_augmentation": False, # this feature is experimental, can be heavily accelerated using rapids_acceleration=True during class instantiation\n            "delete_unpredictable_training_rows": False, # this feature is experimental!\n            "scale_data": False,\n            "smote": False,\n            "autoencoder_based_oversampling": False, # perfect for imbalanced binary and multiclass data\n            "final_pca_dimensionality_reduction": False\n        }\n\nThe bruteforce_random_feature_selection step is experimental. It showed promising results. The number of trials can be controlled.\nThis step is useful, if the model overfitted (which should happen rarely), because too many features with too little\nfeature importance have been considered.\nlike test_class.hyperparameter_tuning_rounds["bruteforce_random"] = 400 .\n\nGenerally the class instance is a control center and gives room for plenty of customization.\nNever update the class attributes like shown below.\n\ntest_class.tabnet_settings = "batch_size": rec_batch_size,\n                                "virtual_batch_size": virtual_batch_size,\n                                # pred batch size?\n                                "num_workers": 0,\n                                "max_epochs": 1000}\n\ntest_class.hyperparameter_tuning_rounds = {"xgboost": 100,\n                                             "lgbm": 500,\n                                             "tabnet": 25,\n                                             "ngboost": 25,\n                                             "sklearn_ensemble": 10,\n                                             "ridge": 500,\n                                             "elasticnet": 100,\n                                             "catboost": 25,\n                                             "sgd": 2000,\n                                             "svm": 50,\n                                             "svm_regression": 50,\n                                             "ransac": 50,\n                                             "multinomial_nb": 100,\n                                             "bruteforce_random": 400,\n                                             "synthetic_data_augmentation": 100,\n                                             "autoencoder_based_oversampling": 200,\n                                             "final_kernel_pca_dimensionality_reduction": 50,\n                                             "final_pca_dimensionality_reduction": 50}\n\ntest_class.hyperparameter_tuning_max_runtime_secs = {"xgboost": 2*60*60,\n                                                       "lgbm": 2*60*60,\n                                                       "tabnet": 2*60*60,\n                                                       "ngboost": 2*60*60,\n                                                       "sklearn_ensemble": 2*60*60,\n                                                       "ridge": 2*60*60,\n                                                       "elasticnet": 2*60*60,\n                                                       "catboost": 2*60*60,\n                                                       "sgd": 2*60*60,\n                                                       "svm": 2*60*60,\n                                                       "svm_regression": 2*60*60,\n                                                       "ransac": 2*60*60,\n                                                       "multinomial_nb": 2*60*60,\n                                                       "bruteforce_random": 2*60*60,\n                                                       "synthetic_data_augmentation": 1*60*60,\n                                                       "autoencoder_based_oversampling": 2*60*60,\n                                                       "final_kernel_pca_dimensionality_reduction": 4*60*60,\n                                                       "final_pca_dimensionality_reduction": 2*60*60}\n\nWhen these parameters have to updated, please overwrite the keys individually to not break the blueprints eventually.\nI.e.: test_class.hyperparameter_tuning_max_runtime_secs["xgboost"] = 12*60*60 would work fine.\n\nWorking with big data can bring all hardware to it\'s needs. e2eml has been tested with:\n- Ryzen 5950x (16 cores CPU)\n- Geforce RTX 3090 (24GB VRAM)\n- 64GB RAM\ne2eml has been able to process 100k rows with 200 columns approximately using these specs stable for non-blended\nblueprints. Blended blueprints consume more resources as e2eml keep the trained models in memory as of now.\n\nFor data bigger than 100k rows it is possible to limit the amount of data for various preprocessing steps:\n- test_class.feature_selection_sample_size = 100000 # for feature selection\n- test_class.hyperparameter_tuning_sample_size = 100000 # for model hyperparameter optimization\n- test_class.brute_force_selection_sample_size = 15000 # for an experimental feature selection\n\nFor binary classification a sample size of 100k datapoints is sufficient in most cases. Hyperparameter tuning sample size can be much less,\ndepending on class imbalance.\n\nFor multiclass we recommend to start with small samples as algorithms like Xgboost and LGBM will easily grow in memory consumption\nwith growing number of classes.\n\nWhenever classes are imbalanced (binary & multiclass) we recommend to use the preprocessing step "autoencoder_based_oversampling".\n"""\n# After running the blueprint the pipeline is done. I can be saved with:\nsave_to_production(test_class, file_name=\'automl_instance\')\n\n# The blueprint can be loaded with\nloaded_test_class = load_for_production(file_name=\'automl_instance\')\n\n# predict on new data (in this case our holdout) with loaded blueprint\nloaded_test_class.ml_bp01_multiclass_full_processing_xgb_prob(holdout_df)\n\n# predictions can be accessed via a class attribute\nprint(churn_class.predicted_classes[\'xgboost\'])\n```\n\n## Linting and Pre-Commit\n\nThis project uses pre-commit to enforce style.\n\nTo install the pre-commit hooks, first install pre-commit into the project\'s\nvirtual environment:\n\n```sh\npip install pre-commit\n```\n\nThen install the project hooks:\n\n```sh\npre-commit install\n```\n\nNow, whenever you make a commit, the linting and autoformatting will\nautomatically run.\n\n## Disclaimer\n\ne2e is not designed to quickly iterate over several algorithms and suggest you\nthe best. It is made to deliver state-of-the-art performance as ready-to-go\nblueprints. e2e-ml blueprints contain:\n\n* preprocessing (outlier, rare feature, datetime, categorical and NLP handling)\n* feature creation (binning, clustering, categorical and NLP features)\n* automated feature selection\n* model training (with crossfold validation)\n* automated hyperparameter tuning\n* model evaluation\n\nThis comes at the cost of runtime. Depending on your data we recommend strong\nhardware.\n\n## Development\n\nThis project uses [poetry](https://python-poetry.org/).\n\nTo install the project for development, run:\n\n```sh\npoetry install\n```\n\nThis will install all dependencies and development dependencies into a virtual\nenvironment.\n\n### Adding or Removing Dependencies\n\nTo add or remove a dependency, use `poetry add <package>` or\n`poetry remove <package>` respectively. Use the `--dev` flag for development\ndependencies.\n\n### Building and Publishing\n\nTo build and publish the project, run\n\n```sh\npoetry publish --build\n```\n\n### Documentation\n\nThis project comes with documentation. To build the docs, run:\n\n```sh\ncd docs\nmake docs\n```\n\nYou may then browse the HTML docs at `docs/build/docs/index.html`.\n\n### Pull Requests\n\nWe welcome Pull Requests! Please make a PR against the `develop` branch.\n\n## Release History\n\n* 2.10.03\n  * Adjusted dependencies for Pandas, Spacy, Optuna, Setuptools, Transformers\n* 2.10.01\n  * Added references & citations to Readme\n  * Added is_imbalanced flag to Timewalk\n  * Removed babel from dependencies & updated some of them\n* 2.9.96\n  * Timewalk got adjustments\n  * Fixed a bug where row deletion has been incompatible with Tabnet\n* 2.9.95\n  * SHAP based feature selection increased to 20 folds (from 10)\n  * less unnecessary print outs\n* 2.9.93\n  * Added SHAP based feature selection\n  * Removed Xgboost from Timewalk as default due to computational and runtime costs\n  * Suppress all warnings of LGBM focal during multiclass tasks\n* 2.9.92\n  * e2eml uses poetry\n  * introduction of Github actions to check linting\n  * bug fix of LGBM focal failing due to missing hyperparameter tuning specifications\n  * preparation for Readthedocs implementation\n* 2.9.9\n  * Added Multinomial Bayes Classifier\n  * Added SVM for regression\n  * Refined Sklearn ensembles\n* 2.9.8\n  * Added Quadrant Discriminent Analysis\n  * Added Support Vector machines\n  * Added Ransac regressor\n* 2.9.7\n  * updated Plotly dependency to 5.4.0\n  * Improved Xgboost for imbalanced data\n* 2.9.6\n  * Added TimeTravel and timewalk: TimeTravel will save the class instance after\n    each preprocessing step, timewalk will automatically try different\n    preprocessing steps with different algorithms to find the best combination\n  * Updated dependencies to use newest versions of scikit-learn and\n    category-encoders\n* 2.9.0\n  * bug fixes with synthetic data augmentation for regression\n  * bug fix of target encoding during regression\n  * enhanced hyperparameter space for autoencoder based oversampling\n  * added final PCA dimensionality reduction as optional preprocessing step\n* 2.8.1\n  * autoencoder based oversampling will go through hyperprameter tuning first\n    (for each class individually)\n  * optimized TabNet performance\n* 2.7.5\n  * added oversampling based on variational autoencoder (experimental)\n* 2.7.4\n  * fixed target encoding for multiclass classification\n  * improved performance on multiclass tasks\n  * improved Xgboost & TabNet performance on binary classification\n  * added auto-tuned clustering as a feature\n* 2.6.3\n  * small bugfixes\n* 2.6.1\n  * Hyperparameter tuning does happen on a sample of the train data from now on\n    (sample size can be controlled)\n  * An experimental feature has been added, which tries to find unpredictable\n    training data rows to delete them from the training (this accelerates\n    training, but costs a bit model performance)\n  * Blueprints can be accelerated with Nvidia RAPIDS (works on clustering only f\n    or now)\n* 2.5.9\n  * optimized loss function for TabNet\n* 2.5.1\n  * Optimized loss function for synthetic data augmentation\n  * Adjusted library dependencies\n  * Improved target encoding\n* 2.3.1\n  * Changed feature selection backend from Xgboost to LGBM\n  * POS tagging is off on default from this version\n* 2.2.9\n  * bug fixes\n  * added an experimental feature to optimize training data with synthetic data\n  * added optional early feature selection (numeric only)\n* 2.2.2\n  * transformers can be loaded into Google Colab from Gdrive\n* 2.1.2\n  * Improved TFIDF vectorizer performance & non transformer NLP applications\n  * Improved POS tagging stability\n* 2.1.1\n  * Completely overworked preprocessing setup (changed API). Preprocessing\n    blueprints can be customized through a class attribute now\n  * Completely overworked special multimodel blueprints. The paricipating\n    algorithms can be customized through a class attribute now\n  * Improved NULL handling & regression performance\n  * Added Catboost & Elasticnet\n  * Updated Readme\n  * First unittests\n  * Added Stochastic Gradient classifier & regressor\n* 1.8.2\n  * Added Ridge classifier and regression as new blueprints\n* 1.8.1\n  * Added another layer of feature selection\n* 1.8.0\n  * Transformer padding length will be max text length + 20% instead of static\n    300\n  * Transformers use AutoModelForSequenceClassification instead of hardcoded\n    transformers now\n  * Hyperparameter tuning rounds and timeout can be controlled globally via\n    class attribute now\n* 1.7.8\n  * Instead of a global probability threshold, e2eml stores threshold for each\n    tested model\n  * Deprecated binary boosting blender due to lack of performance\n  * Added filling of inf values\n* 1.7.3\n  * Improved preprocessing\n  * Improved regression performance\n  * Deprecated regression boosting blender and replaced my multi\n    model/architecture blender\n  * Transformers can optionally discard worst models, but will keep all 5 by\n    default\n  * e2eml should be installable on Amazon Sagemaker now\n* 1.7.0\n  * Added TabNet classifier and regressor with automated hyperparameter\n    optimization\n* 1.6.5\n  * improvements of NLP transformers\n* 1.5.8\n  * Fixes bug around preprocessing_type=\'nlp\'\n  * replaced pickle with dill for saving and loading objects\n* 1.5.3\n  * Added transformer blueprints for NLP classification and regression\n  * renamed Vowpal Wabbit blueprint to fit into blueprint naming convention\n  * Created "extras" options for library installation: \'rapids\' installs extras,\n    so e2eml can be installed into into a rapids environment while \'jupyter\'\n    adds jupyter core and ipython. \'full\' installs all of them.\n* 1.3.9\n  * Fixed issue with automated GPU-acceleration detection and flagging\n  * Fixed avg regression blueprint where eval function tried to call\n    classification evaluation\n  * Moved POS tagging + PCA step into non-NLP pipeline as it showed good results\n    in general\n  * improved NLP part (more and better feature engineering and preprocessing) of\n    blueprints for better performance\n  * Added Vowpal Wabbit for classification and regression and replaced stacking\n    ensemble in automated model exploration by Vowpal Wabbit as well\n  * Set random_state for train_test splits for consistency\n  * Fixed sklearn dependency to 0.22.0 due to six import error\n* 1.0.1\n  * Optimized package requirements\n  * Pinned LGBM requirement to version 3.1.0 due to the bug "LightGBMError: bin\n    size 257 cannot run on GPU #3339"\n* 0.9.9\n  * Enabled tune_mode parameter during class instantiation.\n  * Updated docstings across all functions and changed model defaults.\n  * Multiple bug fixes (LGBM regression accurate mode, label encoding and\n    permutation tests).\n  * Enhanced user information & better ROC_AUC display\n  * Added automated GPU detection for LGBM and Xgboost.\n  * Added functions to save and load blueprints\n  * architectural changes (preprocessing organized in blueprints as well)\n* 0.9.4\n  * First release with classification and regression blueprints. (not available\n    anymore)\n\n## References\n\n* Focal loss\n  * [Focal loss for LGBM](https://maxhalford.github.io/blog/lightgbm-focal-loss/#first-order-derivative)\n  * [Focal loss for LGBM multiclass](https://towardsdatascience.com/multi-class-classification-using-focal-loss-and-lightgbm-a6a6dec28872)\n* Autoencoder\n  * [Variational Autoencoder for imbalanced data](https://github.com/lschmiddey/Autoencoder/blob/master/VAE_for_imbalanced_data.ipynb)\n* Target Encoding\n  * [Target encoding for multiclass](https://towardsdatascience.com/target-encoding-for-multi-class-classification-c9a7bcb1a53)\n* Pytorch-TabNet\n  * [Arik, S. O., & Pfister, T. (2019). TabNet: Attentive Interpretable Tabular Learning. arXiv preprint arXiv:1908.07442.](https://arxiv.org/pdf/1908.07442.pdf)\n  * [Implementing TabNet in Pytorch](https://towardsdatascience.com/implementing-tabnet-in-pytorch-fc977c383279)\n* Ngboost\n  * [NGBoost: Natural Gradient Boosting for Probabilistic Prediction, arXiv:1910.03225](https://arxiv.org/abs/1910.03225)\n* Vowpal Wabbit\n  * [Vowpal Wabbit Research overview](https://vowpalwabbit.org/research.html)\n\n## Meta\n\nCreator: Thomas Meißner – [LinkedIn](https://www.linkedin.com/in/thomas-mei%C3%9Fner-m-a-3808b346)\n\nConsultant: Gabriel Stephen Alexander – [Github](https://github.com/bitsofsteve)\n\nSpecial thanks to: Alex McKenzie - [LinkedIn](https://de.linkedin.com/in/alex-mckenzie)\n\n[e2eml Github repository](https://github.com/ThomasMeissnerDS/e2e_ml)\n',
    'author': 'Thomas Meißner',
    'author_email': 'meissnercorporation@gmx.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ThomasMeissnerDS/e2e_ml',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<3.9',
}


setup(**setup_kwargs)
